@@ -24,6 +24,8 @@
 
 #define GNOME_DESKTOP_USE_UNSTABLE_API
 
+#include <config.h>
+#include <glib/gi18n-lib.h>
 #include <stdlib.h>
 #include <string.h>
 #include <glib.h>
@@ -66,9 +68,11 @@ static gboolean parse_file_gmarkup (const gchar *file,
 
 typedef struct CrtcAssignment CrtcAssignment;
 
-static gboolean         crtc_assignment_apply (CrtcAssignment   *assign);
+static gboolean         crtc_assignment_apply (CrtcAssignment   *assign,
+					       GError          **error);
 static CrtcAssignment  *crtc_assignment_new   (GnomeRRScreen    *screen,
-					       GnomeOutputInfo **outputs);
+					       GnomeOutputInfo **outputs,
+					       GError          **error);
 static void             crtc_assignment_free  (CrtcAssignment   *assign);
 static void             output_free           (GnomeOutputInfo  *output);
 static GnomeOutputInfo *output_copy           (GnomeOutputInfo  *output);
@@ -187,7 +191,7 @@ handle_start_element (GMarkupParseContext *context,
 	
 	parser->configuration = g_new0 (GnomeRRConfig, 1);
 	parser->configuration->clone = FALSE;
-	parser->configuration->outputs = g_new0 (GnomeOutputInfo *, 1);
+	parser->configuration->outputs = NULL;
     }
     else if (strcmp (name, "monitors") == 0)
     {
@@ -657,6 +661,8 @@ outputs_free (GnomeOutputInfo **outputs)
 
     for (i = 0; outputs[i] != NULL; ++i)
 	output_free (outputs[i]);
+
+    g_free (outputs);
 }
 
 void
@@ -746,6 +752,42 @@ output_match (GnomeOutputInfo *output1, GnomeOutputInfo *output2)
     return TRUE;
 }
 
+static gboolean
+output_equal (GnomeOutputInfo *output1, GnomeOutputInfo *output2)
+{
+    g_assert (output1 != NULL);
+    g_assert (output2 != NULL);
+
+    if (!output_match (output1, output2))
+	return FALSE;
+
+    if (output1->on != output2->on)
+	return FALSE;
+
+    if (output1->on)
+    {
+	if (output1->width != output2->width)
+	    return FALSE;
+	
+	if (output1->height != output2->height)
+	    return FALSE;
+	
+	if (output1->rate != output2->rate)
+	    return FALSE;
+	
+	if (output1->x != output2->x)
+	    return FALSE;
+	
+	if (output1->y != output2->y)
+	    return FALSE;
+	
+	if (output1->rotation != output2->rotation)
+	    return FALSE;
+    }
+
+    return TRUE;
+}
+
 static GnomeOutputInfo *
 find_output (GnomeRRConfig *config, const char *name)
 {
@@ -762,6 +804,9 @@ find_output (GnomeRRConfig *config, const char *name)
     return NULL;
 }
 
+/* Match means "these configurations apply to the same hardware
+ * setups"
+ */
 gboolean
 gnome_rr_config_match (GnomeRRConfig *c1, GnomeRRConfig *c2)
 {
@@ -780,6 +825,28 @@ gnome_rr_config_match (GnomeRRConfig *c1, GnomeRRConfig *c2)
     return TRUE;
 }
 
+/* Equal means "the configurations will result in the same
+ * modes being set on the outputs"
+ */
+gboolean
+gnome_rr_config_equal (GnomeRRConfig  *c1,
+		       GnomeRRConfig  *c2)
+{
+    int i;
+
+    for (i = 0; c1->outputs[i] != NULL; ++i)
+    {
+	GnomeOutputInfo *output1 = c1->outputs[i];
+	GnomeOutputInfo *output2;
+
+	output2 = find_output (c2, output1->name);
+	if (!output2 || !output_equal (output1, output2))
+	    return FALSE;
+    }
+    
+    return TRUE;
+}
+
 static GnomeOutputInfo **
 make_outputs (GnomeRRConfig *config)
 {
@@ -820,12 +887,19 @@ make_outputs (GnomeRRConfig *config)
 
 gboolean
 gnome_rr_config_applicable (GnomeRRConfig  *configuration,
-			  GnomeRRScreen       *screen)
+			    GnomeRRScreen  *screen,
+			    GError        **error)
 {
-    GnomeOutputInfo **outputs = make_outputs (configuration);
-    CrtcAssignment *assign = crtc_assignment_new (screen, outputs);
+    GnomeOutputInfo **outputs;
+    CrtcAssignment *assign;
     gboolean result;
 
+    g_return_val_if_fail (configuration != NULL, FALSE);
+    g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+
+    outputs = make_outputs (configuration);
+    assign = crtc_assignment_new (screen, outputs, error);
+
     if (assign)
     {
 	result = TRUE;
@@ -841,21 +915,6 @@ gnome_rr_config_applicable (GnomeRRConfig  *configuration,
     return result;
 }
 
-static GnomeRRConfig *
-gnome_rr_config_find (GnomeRRConfig **haystack,
-		    GnomeRRConfig  *needle)
-{
-    int i;
-
-    for (i = 0; haystack[i] != NULL; ++i)
-    {
-	if (gnome_rr_config_match (haystack[i], needle))
-	    return haystack[i];
-    }
-
-    return NULL;
-}
-
 /* Database management */
 
 static gchar *
@@ -992,14 +1051,19 @@ gnome_rr_config_sanitize (GnomeRRConfig *config)
 
 
 gboolean
-gnome_rr_config_save (GnomeRRConfig *configuration, GError **err)
+gnome_rr_config_save (GnomeRRConfig *configuration, GError **error)
 {
     GnomeRRConfig **configurations;
-    GString *output = g_string_new("");
+    GString *output;
     int i;
     gchar *filename;
     gboolean result;
 
+    g_return_val_if_fail (configuration != NULL, FALSE);
+    g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+
+    output = g_string_new ("");
+
     configurations = configurations_read (NULL); /* NULL-GError */
     
     g_string_append_printf (output, "<monitors version=\"1\">\n");
@@ -1020,7 +1084,7 @@ gnome_rr_config_save (GnomeRRConfig *configuration, GError **err)
     g_string_append_printf (output, "</monitors>\n");
 
     filename = get_config_filename ();
-    result = g_file_set_contents (filename, output->str, -1, err);
+    result = g_file_set_contents (filename, output->str, -1, error);
     g_free (filename);
 
     if (result)
@@ -1037,22 +1101,82 @@ gnome_rr_config_save (GnomeRRConfig *configuration, GError **err)
     return result;
 }
 
-static gboolean
-apply_configuration (GnomeRRConfig *conf, GnomeRRScreen *screen)
+static GnomeRRConfig *
+gnome_rr_config_copy (GnomeRRConfig *config)
+{
+    GnomeRRConfig *copy = g_new0 (GnomeRRConfig, 1);
+    int i;
+    GPtrArray *array = g_ptr_array_new ();
+
+    copy->clone = config->clone;
+    
+    for (i = 0; config->outputs[i] != NULL; ++i)
+	g_ptr_array_add (array, output_copy (config->outputs[i]));
+
+    g_ptr_array_add (array, NULL);
+    copy->outputs = (GnomeOutputInfo **)g_ptr_array_free (array, FALSE);
+
+    return copy;
+}
+
+GnomeRRConfig *
+gnome_rr_config_new_stored (GnomeRRScreen *screen, GError **error)
+{
+    GnomeRRConfig *current;
+    GnomeRRConfig **configs;
+    GnomeRRConfig *result;
+
+    g_return_val_if_fail (screen != NULL, NULL);
+    g_return_val_if_fail (error == NULL || *error == NULL, NULL);
+    
+    current = gnome_rr_config_new_current (screen);
+    
+    configs = configurations_read (error);
+
+    result = NULL;
+    if (configs)
+    {
+	int i;
+	
+	for (i = 0; configs[i] != NULL; ++i)
+	{
+	    if (gnome_rr_config_match (configs[i], current))
+	    {
+		result = gnome_rr_config_copy (configs[i]);
+		break;
+	    }
+	}
+
+	if (result == NULL)
+	    g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_NO_MATCHING_CONFIG,
+			 _("none of the saved display configurations matched the active configuration"));
+
+	configurations_free (configs);
+    }
+
+    gnome_rr_config_free (current);
+    
+    return result;
+}
+
+gboolean
+gnome_rr_config_apply (GnomeRRConfig *config,
+		       GnomeRRScreen *screen,
+		       GError       **error)
 {
     CrtcAssignment *assignment;
     GnomeOutputInfo **outputs;
     gboolean result = FALSE;
 
-    outputs = make_outputs (conf);
+    outputs = make_outputs (config);
 
-    assignment = crtc_assignment_new (screen, outputs);
+    assignment = crtc_assignment_new (screen, outputs, error);
 
     outputs_free (outputs);
     
     if (assignment)
     {
-	if (crtc_assignment_apply (assignment))
+	if (crtc_assignment_apply (assignment, error))
 	    result = TRUE;
 	    
 	crtc_assignment_free (assignment);
@@ -1064,44 +1188,42 @@ apply_configuration (GnomeRRConfig *conf, GnomeRRScreen *screen)
 }
 
 gboolean
-gnome_rr_config_apply_stored (GnomeRRScreen *screen)
+gnome_rr_config_apply_stored (GnomeRRScreen *screen, GError **error)
 {
-    GnomeRRConfig **configs;
-    GnomeRRConfig *current;
-    GnomeRRConfig *found;
-    gboolean result = TRUE;
+    GnomeRRConfig *stored;
+    GError *my_error;
 
-    if (!screen)
-	return FALSE;
+    g_return_val_if_fail (screen != NULL, FALSE);
+    g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
-    configs = configurations_read (NULL); /* NULL-GError */
+    my_error = NULL;
+    if (!gnome_rr_screen_refresh (screen, &my_error)) {
+	    if (my_error) {
+		    g_propagate_error (error, my_error);
+		    return FALSE; /* This is a genuine error */
+	    }
 
-    gnome_rr_screen_refresh (screen);
-    
-    current = gnome_rr_config_new_current (screen);
-
-    if (configs)
-    {
-	if ((found = gnome_rr_config_find (configs, current)))
-	{
-	    apply_configuration (found, screen);
-
-	    result = TRUE;
-	}
-	else
-	{
-	    result = FALSE;
-	}
-	
-	configurations_free (configs);
+	    /* This means the screen didn't change, so just proceed */
     }
+
+    stored = gnome_rr_config_new_stored (screen, error);
+
+    if (stored)
+    {
+	gboolean result;
+
+	result = gnome_rr_config_apply (stored, screen, error);
+
+	gnome_rr_config_free (stored);
 	
-    gnome_rr_config_free (current);
-
-    return result;
+	return result;
+    }
+    else
+    {
+	return FALSE;
+    }
 }
 
-
 /*
  * CRTC assignment
  */
@@ -1214,6 +1336,11 @@ crtc_assignment_free (CrtcAssignment *assign)
     g_free (assign);
 }
 
+typedef struct {
+    gboolean has_error;
+    GError **error;
+} ConfigureCrtcState;
+
 static void
 configure_crtc (gpointer key,
 		gpointer value,
@@ -1221,13 +1348,19 @@ configure_crtc (gpointer key,
 {
     GnomeRRCrtc *crtc = key;
     CrtcInfo *info = value;
+    ConfigureCrtcState *state = data;
 
-    gnome_rr_crtc_set_config (crtc,
-			info->x, info->y,
-			info->mode,
-			info->rotation,
-			(GnomeRROutput **)info->outputs->pdata,
-			info->outputs->len);
+    if (state->has_error)
+	return;
+
+    if (!gnome_rr_crtc_set_config (crtc,
+				   info->x, info->y,
+				   info->mode,
+				   info->rotation,
+				   (GnomeRROutput **)info->outputs->pdata,
+				   info->outputs->len,
+				   state->error))
+	state->has_error = TRUE;
 }
 
 static gboolean
@@ -1368,7 +1501,7 @@ get_required_virtual_size (CrtcAssignment *assign, int *width, int *height)
 }
 
 static CrtcAssignment *
-crtc_assignment_new (GnomeRRScreen *screen, GnomeOutputInfo **outputs)
+crtc_assignment_new (GnomeRRScreen *screen, GnomeOutputInfo **outputs, GError **error)
 {
     CrtcAssignment *assignment = g_new0 (CrtcAssignment, 1);
 
@@ -1388,6 +1521,15 @@ crtc_assignment_new (GnomeRRScreen *screen, GnomeOutputInfo **outputs)
 	if (width < min_width || width > max_width ||
 	    height < min_height || height > max_height)
 	{
+	    g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_BOUNDS_ERROR,
+			 /* Translators: the "requested", "minimum", and
+			  * "maximum" words here are not keywords; please
+			  * translate them as usual. */
+			 _("required virtual size does not fit available size: "
+			   "requested=(%d, %d), minimum=(%d, %d), maximum=(%d, %d)"),
+			 width, height,
+			 min_width, min_height,
+			 max_width, max_height);
 	    goto fail;
 	}
 
@@ -1395,6 +1537,9 @@ crtc_assignment_new (GnomeRRScreen *screen, GnomeOutputInfo **outputs)
 	
 	return assignment;
     }
+    else
+	g_set_error (error, GNOME_RR_ERROR, GNOME_RR_ERROR_CRTC_ASSIGNMENT,
+		     _("could not find a suitable configuration of screens"));
     
 fail:
     crtc_assignment_free (assignment);
@@ -1403,7 +1548,7 @@ fail:
 }
 
 static gboolean
-crtc_assignment_apply (CrtcAssignment *assign)
+crtc_assignment_apply (CrtcAssignment *assign, GError **error)
 {
     GnomeRRCrtc **all_crtcs = gnome_rr_screen_list_crtcs (assign->screen);
     int width, height;
@@ -1425,6 +1570,8 @@ crtc_assignment_apply (CrtcAssignment *assign)
     width = MIN (max_width, width);
     height = MAX (min_height, height);
     height = MIN (max_height, height);
+
+    /* FMQ: do we need to check the sizes instead of clamping them? */
     
     /* Turn off all crtcs that are currently displaying outside the new screen,
      * or are not used in the new setup
@@ -1452,7 +1599,7 @@ crtc_assignment_apply (CrtcAssignment *assign)
 	    
 	    if (x + w > width || y + h > height || !g_hash_table_lookup (assign->info, crtc))
 	    {
-		if (!gnome_rr_crtc_set_config (crtc, 0, 0, NULL, GNOME_RR_ROTATION_0, NULL, 0))
+		if (!gnome_rr_crtc_set_config (crtc, 0, 0, NULL, GNOME_RR_ROTATION_0, NULL, 0, error))
 		{
 		    success = FALSE;
 		    break;
@@ -1473,9 +1620,16 @@ crtc_assignment_apply (CrtcAssignment *assign)
 
     if (success)
     {
+	ConfigureCrtcState state;
+
 	gnome_rr_screen_set_size (assign->screen, width, height, width_mm, height_mm);
+
+	state.has_error = FALSE;
+	state.error = error;
 	
-	g_hash_table_foreach (assign->info, configure_crtc, NULL);
+	g_hash_table_foreach (assign->info, configure_crtc, &state);
+
+	success = !state.has_error;
     }
 
     return success;
